from loguru import logger
from rich import print as rprint

from llmbanter.banter.bot_base import BotBase
from llmbanter.library import llm
from llmbanter.library.classes import Response


class LLMBot(BotBase):
    def __init__(
        self,
        version: float,
        name: str,
        system: str,
        opener: str,
        first_bot: bool = False,
        voice: str = "onyx",
        model: str = "gpt-3.5-turbo",
        temperature: float = 1.0,
        debug: bool = False,
    ):
        super().__init__(version, name, system, opener, first_bot, voice, debug)
        self.model = model
        self.temperature = temperature
        self.conversation = [{"role": "system", "content": self.system}]

    def __repr__(self) -> str:
        custom_yaml = " (custom)" if self.is_local_file else ""
        return f"{type(self).__name__} {self.filename}.yaml{custom_yaml} '{self.name}' {self.model}@{self.temperature}"

    def system_message(self) -> str:
        system_messages = [x for x in self.conversation if x["role"] == "system"]
        assert len(system_messages) > 0, "Expected conversation to have been initialized with system message"
        assert len(system_messages) == 1, "Expected conversation to have a single system message"
        first_system_message = system_messages[0]
        return first_system_message["content"]

    def respond_to(self, user_input: str) -> Response:
        assert len(self.conversation) > 0, "Expected conversation to have been initialized with system role"
        if self.first_bot and len(self.conversation) == 1:
            # Include opener in start of conversation (should only apply for the first initiating bot)
            assert (
                self.opener
            ), f"first_bot was True but no opener provided for bot {self.name}. {len(self.conversation)=}, {self.first_bot=}, {self.opener=}"
            self.conversation.append({"role": "assistant", "content": self.opener})

        self.conversation.append({"role": "user", "content": user_input})

        response = llm.get_response(self.clean_name, self.model, self.temperature, self.conversation, self.debug)
        self.conversation.append({"role": "assistant", "content": response.chat_response})

        if not response.cache_hit:
            # TODO: show this have a cached count and a non-cached count to reflect costs for original queries?
            self.total_prompt_tokens += response.prompt_tokens
            self.total_completion_tokens += response.completion_tokens
            self.total_tokens += response.total_tokens
            self.total_chars += len(response.chat_response)

        response.chat_response = response.chat_response.replace(
            "\\n", "\n"
        )  # TODO: do this in BotResponse class perhaps?
        return response

    def cost_estimate_cents(self) -> float:
        # https://openai.com/pricing#language-models (as of Nov 2023)
        def gpt4_8k_price_estimate(prompt_tokens, completion_tokens) -> float:
            return (prompt_tokens / 1000) * 3 + (completion_tokens / 1000) * 6

        def gpt35_4k_price_estimate(prompt_tokens, completion_tokens) -> float:
            return (prompt_tokens / 1000) * 0.1 + (completion_tokens / 1000) * 0.2

        if self.model.startswith("gpt-3.5"):
            return gpt35_4k_price_estimate(self.total_prompt_tokens, self.total_completion_tokens)
        elif self.model.startswith("gpt-4"):
            return gpt4_8k_price_estimate(self.total_prompt_tokens, self.total_completion_tokens)
        else:
            logger.warning(f"Unknown model {self.model}, can't estimate cost.")
            return -1


class HumanInputBot(BotBase):
    def __init__(
        self,
        version: float,
        name: str,
        first_bot: bool = False,
        voice: str = "onyx",
        debug: bool = False,
        multiline: bool = False,
    ):
        system = ""
        opener = ""
        super().__init__(version, name, system, opener, first_bot, voice, debug)

        self.i = 0
        self.multiline = multiline

    def get_opener(self) -> Response:
        assert (
            self.first_bot and self.i == 0
        ), "get_opener() should only be called for first bot at start of conversation."

        if self.multiline:
            lines = []
            try:
                while True:
                    lines.append(input("You (multiline Ctrl+d to end) (opener): "))
            except EOFError:
                pass
            response = "\n".join(lines)
            self.opener = response
            return Response(response)
        else:
            while True:
                rprint("[u][white]Your message (opener):[/white][/u]", end=" ")
                response = input("").strip()
                if response:
                    self.opener = response
                    return Response(response)

    def respond_to(self, user_input: str) -> Response:
        if self.multiline:
            lines = []
            try:
                while True:
                    lines.append(input("You (multiline Ctrl+d to end): "))
            except EOFError:
                pass
            response = "\n".join(lines)
            return Response(response)
        else:
            while True:
                rprint("[u][white]Your message:[/white][/u]", end=" ")
                response = input("").strip()
                if response:
                    return Response(response)

    def is_human(self) -> bool:
        return True

    def cost_estimate_cents(self) -> float:
        return 0


class FixedResponseBot(BotBase):
    def __init__(
        self,
        version: float,
        name: str,
        opener: str,
        response_list: list,
        first_bot: bool = False,
        voice: str = "onyx",
        debug: bool = False,
    ):
        system = ""
        super().__init__(version, name, system, opener, first_bot, voice, debug)

        self.i = 0
        self.temperature = ""
        self.model = ""
        self.conversation = []
        self.response_list = response_list

    def respond_to(self, user_input: str) -> Response:
        if self.first_bot and self.i == 0:
            # Include opener in start of conversation (should only apply for the first initiating bot)
            assert (
                self.opener
            ), f"first_bot was True but no opener provided for bot {self.name}. {self.i=}, {self.first_bot=}, {self.opener=}"
            self.conversation.append({"role": "assistant", "content": self.opener})

        self.i += 1
        response = self.response_list[self.i % len(self.response_list)]
        return Response(response)

    def cost_estimate_cents(self) -> float:
        return 0



import sys

from loguru import logger

from llmbanter.library import env


def configure(remove_existing: bool = True, logfile: str = "./log/app.log"):
    if remove_existing:
        logger.remove()

    stderr_level = env.get("LOG_STDERR_LEVEL", "CRITICAL")
    if stderr_level:
        logger.add(sys.stderr, level=stderr_level)

    logger.add(
        logfile,
        level=env.get("LOG_FILE_LEVEL", "WARNING"),
        rotation=env.get("LOG_FILE_ROTATION", "00:00"),
    )



import typer

from llmbanter.banter.bot_base import BotBase
from llmbanter.banter.bot_pair import BotPair


class Commands:
    command_indicator = "/"

    @classmethod
    def _get_commands(cls) -> list[str]:
        methods = [
            method
            for method in dir(Commands)
            # NOTE: ignore methods starting with _ and ending with _
            if callable(getattr(Commands, method)) and not method.startswith("_") and not method.endswith("_")
        ]
        return methods

    @classmethod
    def _is_valid_command(cls, input_command: str) -> bool:
        if not cls.is_command_(input_command):
            return False

        command = input_command.strip().lstrip(cls.command_indicator)
        commands = cls._get_commands()
        return command in commands

    @classmethod
    def is_command_(cls, response: str) -> bool:
        command = response.strip()
        return command.startswith(cls.command_indicator)

    @classmethod
    def process_command_(cls, response: str, bots: BotPair) -> str:
        if not cls.is_command_(response):
            return ""

        if not cls._is_valid_command(response):
            return f"Unknown command: {response}"

        command = response.strip()
        found_command = getattr(Commands(), command.lstrip(cls.command_indicator))
        command_response = found_command(bots)
        return command_response + "\n"

    def human1(self, bots: BotPair) -> str:
        human_bot = BotBase.get_human_bot()
        human_bot.system = bots.bot1.system
        human_bot.conversation = bots.bot1.conversation
        human_bot.first_bot = bots.bot1.first_bot
        bots.bot1 = human_bot
        return "Switching bot1 to human..."

    def human2(self, bots: BotPair) -> str:
        human_bot = BotBase.get_human_bot()
        human_bot.system = bots.bot2.system
        human_bot.conversation = bots.bot2.conversation
        human_bot.first_bot = bots.bot2.first_bot
        bots.bot2 = human_bot
        return "Switching bot2 to human..."

    def system1(self, bots: BotPair) -> str:
        return f"Bot 1 system:\n{bots.bot1.system_message()}"

    def system2(self, bots: BotPair) -> str:
        return f"Bot 2 system:\n{bots.bot2.system_message()}"

    def conversation1(self, bots: BotPair) -> str:
        return str(bots.bot1.conversation_without_system)

    def conversation2(self, bots: BotPair) -> str:
        return str(bots.bot2.conversation_without_system)

    def debug1(self, bots: BotPair) -> str:
        bots.bot1.debug = not bots.bot1.debug
        return "Debug mode is now " + ("on" if bots.bot1.debug else "off") + " for bot1."

    def debug2(self, bots: BotPair) -> str:
        bots.bot1.debug = not bots.bot1.debug
        return "Debug mode is now " + ("on" if bots.bot2.debug else "off") + " for bot2."

    def help(self, bots: BotPair) -> str:
        methods = [self.command_indicator + command for command in Commands._get_commands()]
        return "List of user commands:\n" + "\n".join(sorted(methods))

    def quit(self, bots: BotPair) -> str:
        raise typer.Exit()



import hashlib
import pickle
from pathlib import Path

from litellm import completion, exceptions
from litellm.utils import ModelResponse
from loguru import logger

# from tenacity import (
#     retry,
#     retry_if_not_exception_type,
#     stop_after_attempt,
#     wait_exponential,
# )
from llmbanter.library import consts
from llmbanter.library.classes import AppUsageException, Response


def log_retry(state):
    msg = (
        f"Tenacity retry {state.fn.__name__}: {state.attempt_number=}, {state.idle_for=}, {state.seconds_since_start=}"
    )
    if state.attempt_number < 1:
        logger.info(msg)
    else:
        logger.exception(msg)


# @retry(
#     wait=wait_exponential(multiplier=2, min=5, max=600),
#     stop=stop_after_attempt(2),
#     before_sleep=log_retry,
#     retry=retry_if_not_exception_type(AppUsageException | exceptions.APIConnectionError),
# )
def _get_response_cached(
    bot_name: str, model: str, temperature: float, messages: list, cache_folder: str = consts.default_cache_folder_llm
) -> tuple[ModelResponse, bool]:
    args = f"{bot_name}-{model}@{temperature}-{messages}"
    args_hash = hashlib.sha256(args.encode()).hexdigest()
    filename_hash = f"{bot_name}_{model}_{temperature}_{args_hash}.pkl"

    folder = Path(cache_folder) / bot_name
    folder.mkdir(parents=True, exist_ok=True)
    filepath = folder / filename_hash

    if filepath.exists():
        with open(filepath, "rb") as f:
            api_response = pickle.load(f)
            cache_hit = True
    else:
        # TODO: handle errors like:
        # BadRequestError: Error code: 400 - {'error': {'message': "This model's maximum context length is 4097 tokens. However, your messages resulted in 12197 tokens. Please reduce the length of the messages.", 'type': 'invalid_request_error', 'param': 'messages', 'code': 'context_length_exceeded'}}
        # APIError: HTTP code 502 from API, Timeout: Request timed out: HTTPSConnectionPool(host='api.openai.com', port=443): Read timed out. (read timeout=600)
        api_response = completion(
            model=model,
            temperature=temperature,
            messages=messages,
        )
        assert api_response.model and api_response.model.startswith(model)
        with open(filepath, "wb") as f:
            pickle.dump(api_response, f)
        cache_hit = False

    assert isinstance(
        api_response, ModelResponse
    ), f"Expected api_response to be a ModelResponse, not {type(api_response)=}"
    return api_response, cache_hit


def get_response(bot_name: str, model: str, temperature: float, messages: list, debug: bool = False) -> Response:
    assert model is not None, f"Expected value for: {model=}"
    assert temperature is not None, f"Expected value for: {temperature=}"
    assert len(messages) > 0, f"Expected value for: {messages=}"
    # TODO: assert messages contain a single {"role": "system"} entry

    try:
        api_response, cache_hit = _get_response_cached(bot_name, model, temperature, messages)
    except exceptions.APIConnectionError as e:
        raise AppUsageException(
            f"APIConnectionError: {e}\n"
            "See https://litellm.vercel.app/docs/providers for additional information on setting LLM provider, keys etc."
        ) from e

    chat_response = api_response.choices[0].message.content  # type: ignore
    total_tokens = int(api_response["usage"]["total_tokens"])
    prompt_tokens = int(api_response["usage"]["prompt_tokens"])
    completion_tokens = int(api_response["usage"]["completion_tokens"])

    return Response(chat_response, total_tokens, prompt_tokens, completion_tokens, cache_hit)



import hashlib
import json
from datetime import datetime
from pathlib import Path

from loguru import logger
from omegaconf import DictConfig
from rich import print as rprint
from rich.console import Console

from llmbanter.banter.bot_base import BotBase
from llmbanter.banter.bot_pair import BotPair
from llmbanter.library import consts
from llmbanter.library.classes import Response
from llmbanter.library.commands import Commands
from llmbanter.library.format import RichTerminalFormatter
from llmbanter.library.sound import Sound


class Conversation:
    def __init__(
        self,
        config: DictConfig,
        debug: bool = False,
    ):
        self.bot1 = config.bot1
        self.bot2 = config.bot2
        self.model1 = config.model1
        self.model2 = config.model2
        self.temperature1 = config.temperature1
        self.temperature2 = config.temperature2
        self.speak = config.speak
        self.show_costs = config.show_costs
        self.debug = debug

    def _pprint(self, text: str):
        rprint(text)
        logger.info(text)

    def _initialise_bots(self) -> BotPair:
        return BotPair(self.bot1, self.bot2, self.model1, self.model2, self.temperature1, self.temperature2)

    def _get_conversation_details(self, bots: BotPair) -> tuple[Path, str, str]:
        transcript_header = f"{bots.bot1.filename} '{bots.bot1.name}' {bots.bot1.model}@{bots.bot1.temperature} <-> {bots.bot2.filename} '{bots.bot2.name}' {bots.bot2.model}@{bots.bot2.temperature}"
        hash = hashlib.md5(transcript_header.encode("utf-8")).hexdigest()[0:7]
        folder = Path(f"./{consts.package_name}_conversations/")
        folder.mkdir(parents=True, exist_ok=True)
        filename = (
            folder / f"{datetime.today().strftime('%Y%m%d-%H%M%S')}_"
            f"{bots.bot1.filename}#{bots.bot1.model}@{bots.bot1.temperature}_"
            f"{bots.bot2.filename}#{bots.bot2.model}@{bots.bot2.temperature}_{hash}.txt"
        )
        return filename, hash, transcript_header

    def _respond_to(self, bot: BotBase, other_response: Response, color: str) -> tuple[Response, float, bool]:
        mp3_cost_cents = 0.0
        mp3_from_cache = True
        response = ""

        # Bot responds to text input
        console = Console()
        spinner = "layer"
        if bot.is_human():
            response = bot.respond_to(other_response.chat_response)
        else:
            with console.status(f"[u][white]{bot.display_name}:[/white][/u]", spinner=spinner, spinner_style=color):
                response = bot.respond_to(other_response.chat_response)

        # Log conversation
        if not bot.is_human():
            self._pprint(
                f"[u][white]{bot.display_name}:[/white][/u] "
                f"{RichTerminalFormatter().format_response(response.chat_response, color)}"
            )

        # Play mp3
        if self.speak and not bot.is_human():
            mp3_file, mp3_cost_cents, mp3_from_cache = Sound.to_mp3(response.chat_response, bot.voice, bot.clean_name)
            Sound.play_mp3(mp3_file)

        return response, mp3_cost_cents, mp3_from_cache

    def _pause_input(self, bots: BotPair):
        while True:
            rprint("[i][bright_black]press ↵ to continue conversation...[/bright_black][/i]", end=" ")
            user_input = input("").strip()
            if Commands.is_command_(user_input):
                command_response = Commands.process_command_(user_input, bots)
                self._pprint(command_response)
            else:
                break

    def _log_conversation(self, bots: BotPair, filename: Path):
        with open(filename.with_suffix(".json"), "w") as f:
            # Write serialised bots object
            j = json.dumps(bots, default=lambda o: o.__dict__, sort_keys=True, indent=4)
            f.write(j)

        with open(filename.with_suffix(".txt"), "w") as f:
            # Write readable conversation
            f.write(f"{bots.bot1.name} talking with {bots.bot2.name}\n")
            f.write(f"{datetime.now().strftime('%Y-%m-%d %H:%M:%S')}\n")
            f.write("\n")
            f.write(f"{bots.bot1}\n")
            f.write(f"{bots.bot2}\n")
            f.write("\n")
            convo_list = [x["role"] + ": " + x["content"] for i, x in enumerate(bots.bot1.conversation_without_system)]
            convo = "\n\n".join(convo_list)
            f.write(convo)
            f.write("\n")

    def start(self):
        if self.model1.startswith("gpt-4") or self.model2.startswith("gpt-4"):
            self._pprint("[red]WARNING: GPT-4 model activated, watch your costs.[/red]")

        bots = self._initialise_bots()
        self._pprint("Conversation set up:")
        self._pprint(f"{bots.bot1=}")
        self._pprint(f"{bots.bot2=}")
        print()
        self._pprint(Commands().help(bots))  # TODO: include a decription with each command
        print()

        i = 1
        response2 = ""
        total_prompt_tokens = 0
        total_completion_tokens = 0
        total_chars = 0
        total_mp3_cents = 0

        # Bot1 opener
        # TODO: remove this and replace with bot1.respond_to() to get opener
        response1 = bots.bot1.get_opener()
        self._pprint(f"Conversation:\n[bright_black]{i}.[/bright_black]")
        self._pprint(f"[u][white]{bots.bot1.display_name}:[/white][/u] [cyan2]{response1.chat_response}[/cyan2]")
        if self.speak and not bots.bot1.is_human():
            mp3_file, total_mp3_cents, mp3_from_cache = Sound.to_mp3(
                response1.chat_response, bots.bot1.voice, bots.bot1.clean_name
            )
            Sound.play_mp3(mp3_file)

        # Start conversation
        mp3_from_cache1, mp3_from_cache2 = False, False
        filename, hash, transcript_header = self._get_conversation_details(bots)
        while True:
            i += 1

            # Bot 2 responds to Bot 1 opener
            while True:
                response2, mp3_cost_cents2, mp3_from_cache2 = self._respond_to(bots.bot2, response1, "magenta1")
                if Commands.is_command_(response2.chat_response):
                    command_response = Commands.process_command_(response2.chat_response, bots)
                    self._pprint(command_response)
                else:
                    total_mp3_cents += mp3_cost_cents2
                    break

            self._log_conversation(bots, filename)

            # Debug info
            total_llm_cents = bots.bot1.cost_estimate_cents() + bots.bot2.cost_estimate_cents()
            total_cents = total_mp3_cents + total_llm_cents
            if self.show_costs:
                total_prompt_tokens = bots.bot1.total_prompt_tokens + bots.bot2.total_prompt_tokens
                total_completion_tokens = bots.bot1.total_completion_tokens + bots.bot2.total_completion_tokens
                total_chars = bots.bot1.total_chars + bots.bot2.total_chars
                self._pprint(
                    "[i][bright_black]"
                    f"{total_prompt_tokens=}, {total_completion_tokens=}, {total_chars=}"
                    f", {total_mp3_cents=:.1f}, {total_llm_cents=:.2f}, {total_cents=:.2f}"
                    f", model1={self.model1}, model2={self.model2}"
                    f"{' (llm from cache)' if response1.cache_hit or response2.cache_hit else ''}"
                    f"{' (mp3 from cache)' if self.speak and (mp3_from_cache2 or mp3_from_cache1) else ''}"
                    "[/bright_black][i]"
                )

            # Pause if both bots are non-human
            if not bots.bot1.is_human() and not bots.bot2.is_human():
                self._pause_input(bots)

            print()
            self._pprint(f"[bright_black]{i}.[/bright_black]")

            # Bot 1 responds to Bot 2
            while True:
                response1, mp3_cost_cents1, mp3_from_cache1 = self._respond_to(bots.bot1, response2, "cyan2")
                if Commands.is_command_(response1.chat_response):
                    command_response = Commands.process_command_(response1.chat_response, bots)
                    self._pprint(command_response)
                else:
                    total_mp3_cents += mp3_cost_cents1
                    break

            self._log_conversation(bots, filename)



import os
from typing import Optional

from dotenv import load_dotenv
from loguru import logger

load_dotenv()


def get(name: str, default: Optional[str] = None) -> str:
    logger.trace(f"env.get called with name '{name}' and default '{default}'")
    if os.getenv(name):
        return os.environ[name]

    if default is None:
        raise Exception(f"{name} environment variable is not set.")

    return default



import pkg_resources

package_name = "llmbanter"
version = pkg_resources.get_distribution(package_name).version
success_code = 0

default_output_folder = "./_output/"
default_cache_folder_mp3 = "./.cache/mp3"
default_cache_folder_llm = "./.cache/llm"

# https://platform.openai.com/docs/api-reference/chat/create
default_llm_model = "gpt-3.5-turbo"
default_llm_temperature = 0.7  # Default 1 as per OpenAI docs
default_llm_top_p = 1  # Default 1 as per OpenAI docs

# Models
# gpt-3.5-turbo (short, older)
# gpt-3.5-turbo-1106 (longer, newer)
# gpt-4-1106-preview



from llmbanter.banter.bot_base import BotBase


class BotPair:
    def __init__(
        self, bot1_name: str, bot2_name: str, model1: str, model2: str, temperature1: float, temperature2: float
    ):
        assert bot1_name, "bot1_name required"
        assert bot2_name, "bot2_name required"
        assert model1, "model1 required"
        assert model2, "model2 required"

        # Initialization
        self.bot1 = BotBase.get_bot(bot1_name)
        self.bot1.model = model1
        if self.bot1.temperature is None:
            self.bot1.temperature = temperature1
        assert isinstance(self.bot1.system, str), f"system must be a string, not {type(self.bot1.system)=}"

        self.bot2 = BotBase.get_bot(bot2_name)
        self.bot2.model = model2
        if self.bot2.temperature is None:
            self.bot2.temperature = temperature2
        assert isinstance(self.bot2.system, str), f"system must be a string, not {type(self.bot1.system)=}"

        # Pair bots
        self.bot1.pair_with(self.bot2)



import contextlib
import hashlib
import os
import time
from pathlib import Path
from typing import Literal

from loguru import logger
from openai import OpenAI

from llmbanter.library import consts

with contextlib.redirect_stdout(None):
    # https://stackoverflow.com/questions/51464455/how-to-disable-welcome-message-when-importing-pygame
    import pygame


class Sound:
    male_voice1 = "male_voice1"
    male_voice2 = "male_voice2"
    female_voice1 = "female_voice1"
    female_voice2 = "female_voice2"

    @staticmethod
    def openai_voice(voice) -> Literal["alloy", "echo", "fable", "onyx", "nova", "shimmer"]:
        if voice == "male_voice1":
            return "onyx"
        elif voice == "male_voice2":
            return "echo"
        elif voice == "female_voice1":
            return "shimmer"
        elif voice == "female_voice2":
            return "nova"
        else:
            raise ValueError(f"voice {voice} is not valid")

    @staticmethod
    def _write_tts_mp3(tts_service, voice, text, filepath):
        char_count = len(text)
        # NOTE: AWS is $0.016 per 1,000 input characters for Neural
        # NOTE: AWS is $0.004 per 1,000 input characters for Standard
        # NOTE: OpenAI is $0.015 per 1,000 input characters
        estimated_cost_cents = char_count * 0.015  # TODO: this is OpenAI cost, not AWS Polly
        warn_length = 800  # 250
        if char_count > warn_length:
            msg = f"Text length ({char_count}) exceeds {warn_length} characters, this may result in costly TTS. Press Ctrl-C to cancel."
            logger.warning(msg)
            input(msg)

        if tts_service == "openai":
            client = OpenAI(api_key=os.environ["OPENAI_API_KEY"])
            response = client.audio.speech.create(model="tts-1", voice=Sound.openai_voice(voice), input=text)
            response.stream_to_file(filepath)
            return estimated_cost_cents

        else:
            raise ValueError(f"tts_service must be 'aws' or 'openai', not {tts_service=}")

    # TODO: Consider integrating with https://github.com/myshell-ai/openvoice
    # TODO: Consider integrating with https://github.com/Vaibhavs10/insanely-fast-whisper
    @staticmethod
    def to_mp3(
        text: str,
        voice: str,
        bot_name: str,
        tts_service: str = "openai",
        delay_on_cache_hit: float = 0.3,
        cache_folder: str = consts.default_cache_folder_mp3,
    ) -> tuple[Path, float, bool]:
        def get_valid_filename(s):
            s = s.replace(" ", "_").strip()
            return "".join(x for x in s if (x.isalnum() or x in "_-"))

        args = f"{text}-{voice}-{bot_name}"
        args_hash = hashlib.sha256(args.encode()).hexdigest()
        short_text = get_valid_filename(text[0:30].lower().replace(" ", "_").strip())
        filename_hash = f"{bot_name}_{tts_service}_{voice}_{short_text}_{args_hash}.mp3"

        folder = Path(cache_folder)
        folder.mkdir(parents=True, exist_ok=True)
        filepath = folder / filename_hash

        if filepath.exists():
            time.sleep(delay_on_cache_hit)
            return filepath, 0, True  # Load from cache

        estimated_cost_cents = Sound._write_tts_mp3(tts_service, voice, text, filepath)
        return filepath, estimated_cost_cents, False

    @staticmethod
    def play_mp3(mp3_filepath):
        pygame.mixer.init()
        pygame.mixer.music.load(mp3_filepath)
        pygame.mixer.music.play()
        while pygame.mixer.music.get_busy():
            continue



import ast

from pygments import highlight
from pygments.formatters import TerminalFormatter
from pygments.lexers import PythonLexer
from rich.text import Text


class RichTerminalFormatter:
    def is_python_code(self, text: str) -> bool:
        try:
            ast.parse(text)
            return True
        except SyntaxError:
            return False

    def format_response(self, text: str, default_color: str) -> str:
        if self.is_python_code(text):
            # TODO: extract any ```python ... ``` code blocks and format them
            text_code = highlight(text, PythonLexer(), TerminalFormatter())
            return str(Text.from_ansi(text_code))  # Prep for rich text printing
        return f"[{default_color}]{text}[/{default_color}]"



import importlib.resources
import os
from abc import ABC, abstractmethod
from pathlib import Path

import yaml
from loguru import logger

from llmbanter.library.classes import AppUsageException, Response


class BotBase(ABC):
    bot_folder = Path("./bots")

    def __init__(
        self,
        version: float,
        name: str,
        system: str,
        opener: str,
        first_bot: bool = False,
        voice: str = "onyx",
        debug: bool = False,
        filename: str = "",
    ):
        self.version = version
        self.name = name
        self.system = system
        self.opener = opener
        self.first_bot = first_bot
        self.filename = filename
        self.voice = voice
        self.debug = debug

        self.temperature = 1.0
        self.model = ""
        self.conversation = []
        self.is_local_file = False
        self.total_tokens = 0
        self.total_prompt_tokens = 0
        self.total_completion_tokens = 0
        self.total_chars = 0

    @abstractmethod
    def respond_to(self, user_input: str) -> Response:
        pass

    @abstractmethod
    def cost_estimate_cents(self) -> float:
        pass

    @property
    def conversation_without_system(self) -> list[dict]:
        return [x for x in self.conversation if x["role"] == "user" or x["role"] == "assistant"]

    def is_human(self) -> bool:
        return False

    def get_opener(self) -> Response:
        return Response(self.opener)

    def pair_with(self, other: "BotBase"):
        def talking_with_statement(name):
            # return "You are talking with " + name.replace("_", " ") + " (only address them by name once, not more)\n"
            return "You are talking with " + name.replace("_", " ") + "\n"

        self.system = talking_with_statement(other.name) + self.system
        other.system = talking_with_statement(self.name) + other.system
        self.first_bot = True
        other.first_bot = False

    def system_message(self) -> str:
        return "Not applicable for this bot type: " + type(self).__name__

    @property
    def display_name(self) -> str:
        return self.name

    @property
    def clean_name(self) -> str:
        return self.name.lower().strip().replace(" ", "-")

    def __repr__(self) -> str:
        custom_yaml = " (custom)" if self.is_local_file else ""
        return f"{type(self).__name__} {self.filename}.yaml{custom_yaml} '{self.name}'"

    @classmethod
    def _get_all_resources_bot_filenames(cls) -> list[Path]:
        suffix = ".yaml"
        files = []
        resources_folder = Path(
            importlib.resources.files("llmbanter")._paths[0]  # type: ignore
        )  # TODO: HACK: Fix the use of _paths[0] and test across lower Python versions.
        for root, _, filenames in os.walk(resources_folder):
            for filename in filenames:
                if filename.endswith(suffix):
                    files.append(Path(os.path.join(root, filename)))
        return files

    @classmethod
    def get_all_valid_bot_names(cls, include_system: bool = False) -> list[str]:
        all_bot_filenames = cls._get_all_resources_bot_filenames()
        filenames = [
            f.parent.name + "/" + f.name.replace(".yaml", "")
            for f in all_bot_filenames
            if (include_system or not f.name.startswith("_")) and not f.parent.name == "testing"
        ]
        filenames = sorted(filenames)
        filenames = [f.replace(f"{cls.bot_folder}/", "") for f in filenames]  # TODO: HACK: make this better
        return filenames

    @classmethod
    def get_human_bot(cls) -> "BotBase":
        return cls.get_bot("human")

    @classmethod
    def _get_bot_yaml_file(cls, bot_folder: str, bot_name: str) -> tuple[Path, bool]:
        if len(bot_name.split("/")) == 2:
            # bot_name format 1: 'bot_type/bot_name'
            bot_path = bot_name.split("/")[0]
            bot_name = bot_name.split("/")[1]
            local_path = Path(bot_folder) / bot_path / f"{bot_name}.yaml"
            resources_file = Path(
                str(importlib.resources.files("llmbanter").joinpath(f"{cls.bot_folder}/{bot_path}/{bot_name}.yaml"))
            )
        elif len(bot_name.split("/")) == 1:
            # bot_name format 2: 'bot_name'
            local_path = Path(bot_folder) / f"{bot_name}.yaml"
            resources_file = Path(
                str(importlib.resources.files("llmbanter").joinpath(f"{cls.bot_folder}/{bot_name}.yaml"))
            )
        else:
            raise AppUsageException(
                f"Bot name should be in the format 'bot_type/bot_name' e.g. 'evangelist/instagram'. You specified bot name of '{bot_name}'."
            )

        if local_path.is_file():
            logger.info(f"Using locally found bot file: '{local_path}'")
            is_local_file = True
            return local_path, is_local_file
        elif resources_file.is_file():
            logger.info(f"Using packaged resources bot file: '{resources_file}'")
            is_local_file = False
            return resources_file, is_local_file

        raise AppUsageException(
            f"Given bot name '{bot_name}' was not found locally at ({local_path}) or in the package resources folder ({resources_file}). "
            f"\nCheck the bot name, or try one of: {', '.join(cls.get_all_valid_bot_names())}"
            "\n"
        )

    @classmethod
    def get_bot(cls, bot_name: str, bot_folder: str = "./bots") -> "BotBase":
        file_path, is_local_file = cls._get_bot_yaml_file(bot_folder, bot_name)
        bot_filename = file_path.name.replace(file_path.suffix, "")

        try:
            with open(file_path) as f:
                data = yaml.safe_load(f)
                bot_type = data.pop("bot_type")
                bot_types_module = importlib.import_module("llmbanter.banter.bot_types")
                DynamicBotClass = getattr(bot_types_module, bot_type)
                found_bot: BotBase = DynamicBotClass(**data)
                found_bot.filename = bot_filename
                found_bot.is_local_file = is_local_file

                return found_bot
        except FileNotFoundError as e:
            raise AppUsageException(
                f"Bot name '{bot_name}' not found at {file_path}. Try one of: {cls.get_all_valid_bot_names()}"
            ) from e
        except TypeError as e:
            raise AppUsageException(
                f"Bot file {file_path} has incorrect parameters set: {e}. Please fix and run again."
            ) from e



from typing import Optional

import typer
from loguru import logger
from omegaconf import OmegaConf
from rich import print
from typing_extensions import Annotated

from llmbanter.banter.conversation import Conversation
from llmbanter.library import consts, env, log
from llmbanter.library.classes import AppUsageException

app = typer.Typer()


def version_callback(value: bool):
    if value:
        print(f"{consts.package_name} version: {consts.version}")
        raise typer.Exit()


@app.command()
def run(
    bot1: Annotated[str, typer.Argument(help="Name of the first bot")],
    bot2: Annotated[str, typer.Argument(help="Name of the second bot")],
    model1: Annotated[str, typer.Argument(help="Model of the first bot")] = consts.default_llm_model,
    model2: Annotated[str, typer.Argument(help="Model of the second bot")] = consts.default_llm_model,
    temperature1: Annotated[
        float, typer.Argument(help="Temperature of the first bot")
    ] = consts.default_llm_temperature,
    temperature2: Annotated[
        float, typer.Argument(help="Temperature of the second bot")
    ] = consts.default_llm_temperature,
    speak: Annotated[bool, typer.Option(help="Enable speaking")] = False,
    show_costs: Annotated[bool, typer.Option(help="Show costs during usage")] = True,
    verbose: Annotated[bool, typer.Option(help="Verbose mode")] = False,
    version: Annotated[
        Optional[bool],
        typer.Option("--version", help=f"Display {consts.package_name} version", callback=version_callback),
    ] = None,
) -> None:
    """
    App entry point
    """
    typer_args = locals()
    config = OmegaConf.create(typer_args)
    example_usage = f"Example usage: [bold green]{consts.package_name} python_language_evangelist java_language_evangelist[/bold green]"

    print(f"[bold green]Welcome to {consts.package_name}![/bold green]")
    print()

    log.configure()
    logger.info(f"Start {consts.package_name}, {config=}")
    if verbose:
        print(f"Start {consts.package_name}, {config=}")

    try:
        expected_typer_args = {
            "bot1",
            "bot2",
            "model1",
            "model2",
            "show_costs",
            "speak",
            "temperature1",
            "temperature2",
            "verbose",
            "version",
        }
        assert (
            typer_args.keys() == expected_typer_args
        ), f"Typer arg list: {list(typer_args.keys())}, was not as expected: {expected_typer_args}"

        open_api_key = env.get("OPENAI_API_KEY", "")
        if (model1 == consts.default_llm_model or model2 == consts.default_llm_model) and open_api_key is None:
            raise AppUsageException(
                f"Expected an environment variable 'OPENAI_API_KEY' to be set to use OpenAI API with model {consts.default_llm_model}."
                "\nSee the OpenAI docs for more info on setting your OPENAI_API_KEY: https://platform.openai.com/docs/quickstart/step-2-setup-your-api-key"
                f"\nAlternatively you can use many other providers via LiteLLM which is used under the hood by {consts.package_name}, see: https://litellm.vercel.app/docs/providers"
                "\n"
            )

        conversation = Conversation(config=config)
        conversation.start()
        raise typer.Exit(consts.success_code)

    except AppUsageException as ex:
        print(f"[bold red]{str(ex)}[/bold red]")
        print(example_usage)
        print("")
        print(f"For more information, try '{consts.package_name} --help'.")
        logger.info(ex)

    except typer.Exit as ex:
        if ex.exit_code == consts.success_code:
            print()
            print(
                f"[bold green]Goodbye and thanks for using {consts.package_name}! Please consider starring the project on github: https://github.com/dylanhogg/{consts.package_name}[/bold green]"
            )
            return
        print(f"[bold red]Unexpected error code: {str(ex)}[/bold red]")
        print(example_usage)
        print("")
        print(f"For more information, try '{consts.package_name} --help'.")
        logger.exception(ex)

    except Exception as ex:
        print(f"[bold red]Unexpected exception: {type(ex)} {str(ex)}[/bold red]")
        print(example_usage)
        print("")
        print(f"For more information, try '{consts.package_name} --help'.")
        logger.exception(ex)


if __name__ == "__main__":
    # Module entrypoint
    # e.g. `poetry run python -m llmbanter.console assistant human`
    app()



from dataclasses import dataclass


class AppUsageException(Exception):
    pass


@dataclass
class Response:
    def __init__(
        self,
        chat_response: str,
        total_tokens: int = 0,
        prompt_tokens: int = 0,
        completion_tokens: int = 0,
        cache_hit: bool = False,
    ):
        self.chat_response = chat_response
        self.total_tokens = total_tokens
        self.prompt_tokens = prompt_tokens
        self.completion_tokens = completion_tokens
        self.cache_hit = cache_hit

    chat_response: str
    total_tokens: int
    prompt_tokens: int
    completion_tokens: int
    cache_hit: bool
